import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.wcs.utils import fit_wcs_from_points

from twirl.geometry import pad
from twirl.match import cross_match, find_transform, get_transform_matrix


def compute_wcs(xy: np.ndarray, radecs: np.ndarray, tolerance: int = 5):
    """Compute WCS based on unordered pixel vs. sky coordinates

    Parameters
    ----------
    xy : np.ndarray
        pixel coordinates
    radecs : np.ndarray
        RA-DEC coordinates (in deg)
    tolerance : int, optional
        minimum distance (in units of xy) between points to be cross-matched,
        by default 5

    Returns
    -------
    astropy.wcs.WCS
        image WCS
    """
    M = find_transform(radecs, xy, tolerance=tolerance)
    radecs_xy = (M @ pad(radecs).T)[0:2].T
    i, j = cross_match(xy, radecs_xy).T
    M = get_transform_matrix(radecs[j], xy[i])
    radecs_xy = (M @ pad(radecs).T)[0:2].T
    i, j = cross_match(xy, radecs_xy).T
    return fit_wcs_from_points(xy[i].T, SkyCoord(radecs[j], unit="deg"))


def gaia_radecs(center, fov, limit=10000, circular=True):
    """
    Return RA-DEC (deg) of gaia stars in the image based on the image center and
    field-of-view


    query from https://gea.esac.esa.int/archive/documentation/GEDR3/Gaia_archive/
    chap_datamodel/sec_dm_main_tables/ssec_dm_gaia_source.html

    TODO: adapt to non-square images

    Parameters
    ----------
    center : tuple or astropy.coordinates.SkyCoord
        image center sky coordinates (deg)
    fov : float or astropy.units.Unit
        field-of-view (deg)
    limit : int, optional
        limit queried sources, by default 10000
    circular : bool, optional
        whether query is circular, by default True

    Returns
    -------
    np.ndarray
        RA-DEC, shape (n, 2)
    """

    from astroquery.gaia import Gaia

    if isinstance(center, SkyCoord):
        ra = center.ra.to(u.deg).value
        dec = center.dec.to(u.deg).value

    if not isinstance(fov, u.Quantity):
        fov = fov * u.deg

    if fov.ndim == 1:
        ra_fov, dec_fov = fov.to(u.deg).value
    else:
        ra_fov = dec_fov = fov.to(u.deg).value

    radius = np.min([ra_fov, dec_fov]) / 2

    fields = "ra, dec"

    if circular:
        job = Gaia.launch_job(
            f"select top {limit} {fields} from gaiadr2.gaia_source where "
            "1=CONTAINS("
            f"POINT('ICRS', {ra}, {dec}), "
            f"CIRCLE('ICRS',ra, dec, {radius}))"
            "order by phot_g_mean_mag"
        )
    else:
        job = Gaia.launch_job(
            f"select top {limit} {fields} from gaiadr2.gaia_source where "
            f"ra BETWEEN {ra-ra_fov/2} AND {ra+ra_fov/2} AND "
            f"dec BETWEEN {dec-dec_fov/2} AND {dec+dec_fov/2} "
            "order by phot_g_mean_mag"
        )

    table = job.get_results()
    return np.array([table["ra"].value.data, table["dec"].value.data]).T



import numpy as np

from twirl.geometry import get_transform_matrix, pad
from twirl.quads import hashes


def count_cross_match(xy1, xy2, tol=1e-3):
    return np.count_nonzero(
        np.min(np.linalg.norm(xy1[:, None, :] - xy2[None, :, :], axis=-1), 0) < tol
    )


def cross_match(xy1, xy2, tolerance=10):
    matches = []

    for i, s in enumerate(xy1):
        distances = np.linalg.norm(s - xy2, axis=1)
        closest = np.argmin(distances)
        if distances[closest] < tolerance:
            matches.append([i, closest])

    return np.array(matches)


def find_transform(
    xy2: np.ndarray, xy1: np.ndarray, tolerance: int = 12, min_match: int = None
) -> np.ndarray:
    """_summary_

    Parameters
    ----------
    xy2 : np.ndarray
        first set of points coordinates, shape (n, 2)
    xy1 : np.ndarray
        second set of points coordinates, shape (m, 2)
    tolerance : int, optional
        tolerance of the match, given in xy1 points units, by default 12
    min_match : _type_, optional
        if None, stops when the number of crossed matched points
        is min_match, by default None

    Returns
    -------
    _type_
        _description_
    """

    hash1, quads1 = hashes(xy1)
    hash2, quads2 = hashes(xy2)
    distances = np.linalg.norm(hash1[:, None, :] - hash2[None, :, :], axis=2)
    shortest_hash = np.argmin(distances, 1)
    ns = []

    for i, j in enumerate(shortest_hash):
        M = get_transform_matrix(quads2[j], quads1[i])
        test = (M @ pad(xy2).T)[0:2].T
        n = count_cross_match(xy1, test, tolerance)
        if min_match is not None:
            if n >= min_match:
                break
        else:
            ns.append(n)

        i = np.argmax(ns)
        M = get_transform_matrix(quads2[np.argmin(distances, 1)[i]], quads1[i])

    return M



from itertools import combinations

import numpy as np

from twirl.geometry import proj, u1u2


def _reorder(q):
    """
    order coordinates from closest to first one
    """
    distances = np.linalg.norm(q[:, :, None] - q.T[None, :, :], axis=1)
    q0i = np.min(np.unravel_index(np.argmax(distances), distances.shape))
    q0 = q[q0i]
    q123 = np.delete(q.copy(), q0i, axis=0)
    distance_from_q0 = np.linalg.norm(q123 - q0, axis=1)
    idxs = np.argsort(1 / distance_from_q0)
    return np.array([q0, *q123[idxs]])


reorder = np.vectorize(_reorder, signature="(n, m)->(n, m)")


def good_quads(quads, circletol=0.01):
    """
    whether all points are contained in a circle (see Lang2009)
    """
    q0, q1 = quads[:, 0], quads[:, 1]
    r = np.linalg.norm(q1 - q0, axis=1) / 2
    center = q0 + (q1 - q0) / 2
    in_circle = np.linalg.norm(quads - center[:, None, :], axis=2) <= r[:, None] * (
        1 + circletol
    )
    return np.all(in_circle, axis=1)


def _quad_hash(quad):
    """
    from 4 coordinates produce the quad hash code
    """
    a, b, c, d = quad
    u1, u2 = u1u2(a, b)
    h = np.linalg.norm(b - a)
    return np.array(
        [
            proj(c, a, u1, norm=True) / h,
            proj(d, a, u1, norm=True) / h,
            proj(c, a, u2, norm=True) / h,
            proj(d, a, u2, norm=True) / h,
        ]
    ), np.array([a, b])


quad_hash = np.vectorize(_quad_hash, signature="(n,m)->(k), (l, 2)")


def clean_quads(xy):
    assert xy.shape[1] == 2
    quads_idxs = np.array(list(combinations(np.arange(xy.shape[0]), 4)))
    quads = xy[quads_idxs]
    ordered_quads = reorder(quads)
    return ordered_quads[good_quads(ordered_quads)]


def hashes(xy):
    assert xy.shape[1] == 2
    quads_idxs = np.array(list(combinations(np.arange(xy.shape[0]), 4)))
    quads = xy[quads_idxs]
    ordered_quads = reorder(quads)
    good_ordered_quads = ordered_quads[good_quads(ordered_quads)]
    h, u = quad_hash(good_ordered_quads)
    # we sort hashes from larger AB (see Lang 2008)
    idxs = np.argsort(np.linalg.norm(u[:, 1] - u[:, 0], axis=1))
    return h[idxs[::-1]], good_ordered_quads[idxs[::-1]]



import numpy as np


def pad(x):
    return np.hstack([x, np.ones((x.shape[0], 1))])


def transform_matrix(scale=1.0, rotation=0.0, translation=(0.0, 0.0)):
    """Affine transformation matix

    Parameters
    ----------
    scale : float, optional
        scaling factor, by default 1.
    rotation : float, optional
        rotation angle in radians, by default 0.
    translation : tuple, optional
        translation vector, by default (0., 0.)

    Returns
    -------
    np.ndarray
        transformation matrix
    """

    R = np.array(
        [
            [np.cos(rotation), np.sin(-rotation), 0],
            [np.sin(rotation), np.cos(rotation), 0],
            [0, 0, 1],
        ]
    )
    T = np.array([[1, 0, translation[0]], [0, 1, translation[1]], [0, 0, 1]])
    S = np.array([[scale, 0, 0], [0, scale, 0], [0, 0, 1]])
    return T @ S @ R


def proj(p, origin, axe, norm=False):
    """
    projection of a point p on a segment from origin to axe
    """
    n = axe - origin
    n /= np.linalg.norm(n, 2)
    return np.where(norm, np.dot(p - origin, n), origin + n * np.dot(p - origin, n))[0]


def rotation(point, pivot, angle, norm=False):
    """
    rotate point around pivot of certain angle
    """
    co = np.cos(angle)
    si = np.sin(angle)
    r = np.array([[co, -si], [si, co]])
    x = r @ (point - pivot)
    x = np.where(norm, (x / np.linalg.norm(x)) * np.linalg.norm(point - pivot) * co, x)
    return x + pivot


def u1u2(a, b, norm=False):
    """
    x, y basis as defined in Lang2009
    """
    norm = np.where(norm, np.linalg.norm(b - a), False)
    x = rotation(b, a, -np.pi / 4, norm=norm)
    y = rotation(b, a, np.pi / 4, norm=norm)
    return x, y


def _get_transform_matrix(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    """Compute the transformation matrix between two vectors defined by pairs of points coordinates

    Parameters
    ----------
    v1 : np.array
        pair of points coordinates, shape (2, 2)
    v2 :  np.array
        pair of points coordinates, shape (2, 2)

    Returns
    -------
    np.ndarray
        transformation matrix of shape (3, 3)
    """
    u1 = v1[1] - v1[0]
    u2 = v2[1] - v2[0]
    n1 = np.linalg.norm(u1)
    n2 = np.linalg.norm(u2)
    theta = np.arccos(np.clip(np.dot(u1 / n1, u2 / n2), -1.0, 1.0))
    scale = n2 / n1
    A = transform_matrix(scale=scale, rotation=theta)
    a = np.pad(v1[0], (0, 1), constant_values=1)
    b = np.pad(v2[0], (0, 1), constant_values=1)
    t = (b - A @ a)[0:2]
    return transform_matrix(scale=scale, rotation=theta, translation=t)


def get_transform_matrix(xy1, xy2):
    XY1 = pad(xy1)
    XY2 = pad(xy2)
    M, _, _, _ = np.linalg.lstsq(XY1, XY2, rcond=None)
    return M.T
